from fastaudio.core.spectrogram import get_usable_kwargs


def test_kwargs(a: int = 10, b: int = 20):
    pass


kwargs = {"a": 1, "b": 2}


def test_get_usable_for_function():
    assert get_usable_kwargs(test_kwargs, kwargs) == kwargs


def test_flatten_kwargs():
    extra_kwargs = {"z": 0, "a": 1, "b": 2, "c": 3}
    assert get_usable_kwargs(test_kwargs, extra_kwargs, []) == kwargs



import colorednoise as cn
import torch
from enum import Enum
from fastai.imports import np, random
from fastai.vision.augment import RandTransform
from fastcore.transform import Transform

from ..core.signal import AudioTensor
from ..core.spectrogram import AudioSpectrogram
from ..util import auto_batch
from .functional import NoiseColor, add_noise_, random_mask, region_mask


class AudioPadType(Enum):
    "All methods of padding audio as attributes to get tab-completion and typo-proofing",
    Zeros = "zeros"
    Zeros_After = "zeros_after"
    Repeat = "repeat"


class ResizeSignal(Transform):
    """Crops signal to be length specified in ms by duration, padding if needed"""

    def __init__(self, duration, pad_mode=AudioPadType.Zeros):
        self.duration = duration
        self.pad_mode = pad_mode
        if pad_mode not in [
            AudioPadType.Zeros,
            AudioPadType.Zeros_After,
            AudioPadType.Repeat,
        ]:
            raise ValueError(
                f"""pad_mode {pad_mode} not currently supported,
                only AudioPadType.Zeros, AudioPadType.Zeros_After,
                or AudioPadType.Repeat"""
            )

    def encodes(self, ai: AudioTensor) -> AudioTensor:
        sig = ai.data
        orig_samples = ai.nsamples
        crop_samples = int((self.duration / 1000) * ai.sr)
        if orig_samples == crop_samples:
            return ai
        elif orig_samples < crop_samples:
            ai.data = _tfm_pad_signal(sig, crop_samples, pad_mode=self.pad_mode)
        else:
            crop_start = random.randint(0, int(orig_samples - crop_samples))
            ai.data = sig[:, crop_start : crop_start + crop_samples]
        return ai


def _tfm_pad_signal(sig, width, pad_mode=AudioPadType.Zeros):
    """Pad spectrogram to specified width, using specified pad mode"""
    c, x = sig.shape
    if pad_mode in [AudioPadType.Zeros, AudioPadType.Zeros_After]:
        zeros_front = (
            random.randint(0, width - x) if pad_mode == AudioPadType.Zeros else 0
        )
        pad_front = torch.zeros((c, zeros_front))
        pad_back = torch.zeros((c, width - x - zeros_front))
        return torch.cat((pad_front, sig, pad_back), 1)
    elif pad_mode == AudioPadType.Repeat:
        repeats = width // x + 1
        return sig.repeat(1, repeats)[:, :width]


def _shift(sig, s):
    if s == 0:
        return sig
    out = torch.zeros_like(sig)
    if s < 0:
        out[..., :s] = sig[..., -s:]
    else:
        out[..., s:] = sig[..., :-s]
    return out


def shift_signal(t: torch.Tensor, shift, roll):
    # refactor 2nd half of this statement to just take and roll the final axis
    if roll:
        t.data = torch.from_numpy(np.roll(t.numpy(), shift, axis=-1))
    else:
        t.data = _shift(t, shift)
    return t


class SignalShifter(RandTransform):
    """Randomly shifts the audio signal by `max_pct` %.
    direction must be -1(left) 0(bidirectional) or 1(right).
    """

    def __init__(self, p=0.5, max_pct=0.2, max_time=None, direction=0, roll=False):
        if direction not in [-1, 0, 1]:
            raise ValueError("Direction must be -1(left) 0(bidirectional) or 1(right)")
        self.max_pct = max_pct
        self.max_time = max_time
        self.direction = direction
        self.roll = roll
        super().__init__(p=p)

    def before_call(self, b, split_idx):
        super().before_call(b, split_idx)
        self.shift_factor = random.uniform(-1, 1)
        if self.direction != 0:
            self.shift_factor = self.direction * abs(self.shift_factor)

    def encodes(self, ai: AudioTensor):
        if self.max_time is None:
            s = self.shift_factor * self.max_pct * ai.nsamples
        else:
            s = self.shift_factor * self.max_time * ai.sr
        ai.data = shift_signal(ai.data, int(s), self.roll)
        return ai

    def encodes(self, sg: AudioSpectrogram):
        if self.max_time is None:
            s = self.shift_factor * self.max_pct * sg.width
        else:
            s = self.shift_factor * self.max_time * sg.sr
        return shift_signal(sg, int(s), self.roll)


class AddNoise(Transform):
    "Adds noise of specified color and level to the audio signal"

    def __init__(self, noise_level=0.05, color=NoiseColor.White):
        self.noise_level = noise_level
        self.color = color
        if color not in [*range(-2, 3)]:
            raise ValueError(f"color {color} is not valid")

    def encodes(self, ai: AudioTensor) -> AudioTensor:
        # if it's white noise, implement our own for speed
        if self.color == NoiseColor.White:
            noise = torch.randn_like(ai.data)
        else:
            noise = torch.from_numpy(
                cn.powerlaw_psd_gaussian(exponent=self.color, size=ai.nsamples)
            ).float()
        scaled_noise = noise * ai.data.abs().mean() * self.noise_level
        ai.data += scaled_noise
        return ai


class AddNoiseGPU(Transform):
    """Adds colored noise to the input audio.

    Works on both `AudioTensor` and `AudioSpectrogram` objects.

    """

    def __init__(self, p=1.0, min_level=0.0, max_level=0.05, color=NoiseColor.White):
        self.min_level = min_level
        self.max_level = max_level
        if not NoiseColor.valid(color):
            raise ValueError(f"color {color} is not valid")
        self.color = color
        self.p = p
        super().__init__()

    def before_call(self, b, split_idx):
        self.do = True

    def _encodes(self, tensor):
        return add_noise_(
            tensor.data, self.color, self.min_level, self.max_level, self.p
        )

    @auto_batch(2)
    def encodes(self, ai: AudioTensor):
        return self._encodes(ai)

    @auto_batch(3)
    def encodes(self, sg: AudioSpectrogram):
        return self._encodes(sg)


class ChangeVolume(RandTransform):
    "Changes the volume of the signal"

    def __init__(self, p=0.5, lower=0.5, upper=1.5):
        self.lower, self.upper = lower, upper
        super().__init__(p=p)

    def before_call(self, b, split_idx):
        super().before_call(b, split_idx)
        self.gain = random.uniform(self.lower, self.upper)

    def encodes(self, ai: AudioTensor):
        return ai.apply_gain(self.gain)


class ChangeVolumeGPU(Transform):
    """Change the volume of the signal."""

    def __init__(self, p=0.5, lower=0.5, upper=1.5):
        self.lower, self.upper = lower, upper
        self.p = p
        super().__init__()

    @auto_batch(2)
    def encodes(self, ai: AudioTensor):
        op_shape = [ai.size(0), 1, 1]
        scales = torch.where(
            random_mask(op_shape, self.p, device=ai.device),
            # Volume scaling
            torch.empty(op_shape, device=ai.device).uniform_(self.lower, self.upper),
            # No scaling
            torch.ones(op_shape, device=ai.device),
        )
        ai *= scales
        return ai


class SignalCutout(RandTransform):
    "Randomly zeros some portion of the signal"

    def __init__(self, p=0.5, max_cut_pct=0.15):
        self.max_cut_pct = max_cut_pct
        super().__init__(p=p)

    def before_call(self, b, split_idx):
        super().before_call(b, split_idx)
        self.cut_pct = random.uniform(0, self.max_cut_pct)

    def encodes(self, ai: AudioTensor):
        return ai.cutout(self.cut_pct)


class SignalCutoutGPU(Transform):
    """Zeros a continuous section of the signal."""

    def __init__(self, p=0.5, min_cut_pct=0.0, max_cut_pct=0.15):
        self.max_cut_pct = max_cut_pct
        self.min_cut_pct = min_cut_pct
        self.p = p
        super().__init__()

    @auto_batch(2)
    def encodes(self, ai: AudioTensor):
        n, c, s = ai.shape

        mask = (
            region_mask(
                n, (s * self.min_cut_pct), (s * self.max_cut_pct), s, device=ai.device
            )
            # Only mask some items in the batch
            * random_mask([n, 1], self.p, device=ai.device)
        )
        # Only apply mask to a random subset of items
        ai.masked_fill_(mask.view(n, 1, s), 0.0)

        return ai


class SignalLoss(RandTransform):
    "Randomly loses some portion of the signal"

    def __init__(self, p=0.5, max_loss_pct=0.15):
        self.max_loss_pct = max_loss_pct
        super().__init__(p=p)

    def before_call(self, b, split_idx):
        super().before_call(b, split_idx)
        self.loss_pct = random.uniform(0, self.max_loss_pct)

    def encodes(self, ai: AudioTensor):
        return ai.lose_signal(self.loss_pct)


class SignalLossGPU(Transform):
    """Randomly loses some percentage of samples (non-continuous).

    The same points will be lost across channels, but different points will be
    lost per-item.

    """

    def __init__(self, p=0.5, min_cut_pct=0.0, max_cut_pct=0.15):
        self.max_cut_pct = max_cut_pct
        self.min_cut_pct = min_cut_pct
        self.p = p
        super().__init__()

    @auto_batch(2)
    def encodes(self, ai: AudioTensor):
        op_shape = [ai.size(0), 1, 1]

        cut_pcts = torch.empty(op_shape, device=ai.device).uniform_(
            self.min_cut_pct, self.max_cut_pct
        )
        masks = (
            random_mask(ai.shape, cut_pcts, device=ai.device)
            # Only mask some items.
            * random_mask(op_shape, self.p, device=ai.device)
        )
        ai.masked_fill_(masks, 0)

        return ai


# downmixMono was removed from torchaudio, we now just take the mean across channels
# this works for both batches and individual items
class DownmixMono(Transform):
    "Transform multichannel audios into single channel"

    def encodes(self, ai: AudioTensor) -> AudioTensor:
        downmixed = ai.data.contiguous().mean(-2).unsqueeze(-2)
        return AudioTensor(downmixed, ai.sr)



import papermill as pm
from glob import glob


def test_can_run_notebooks():
    # Search for all notebooks in directory
    notebooks = glob("**/*.ipynb")
    for nb in notebooks:
        try:
            pm.execute_notebook(
                nb, "/dev/null", progress_bar=False, kernel_name="python"
            )
        except Exception as e:
            # Which notebook caused the error
            raise Exception(nb, e)



import random
import torch
import torchaudio
from collections import OrderedDict
from fastai.data.external import URLs
from fastai.data.transforms import Transform, get_files
from fastai.imports import Path, mimetypes, plt, tarfile
from fastai.torch_core import TensorBase, _fa_rebuild_qtensor, _fa_rebuild_tensor
from fastai.vision.data import get_grid
from fastcore.basics import patch
from fastcore.dispatch import typedispatch
from fastcore.meta import delegates
from fastcore.utils import ifnone
from IPython.display import Audio, display
from librosa.display import waveplot
from os import path

audio_extensions = tuple(
    str.lower(k) for k, v in mimetypes.types_map.items() if v.startswith("audio/")
)


def get_audio_files(path, recurse=True, folders=None):
    "Get audio files in `path` recursively, only in `folders`, if specified."
    return get_files(
        path, extensions=audio_extensions, recurse=recurse, folders=folders
    )


def AudioGetter(suf="", recurse=True, folders=None):
    """Create `get_audio_files` partial function that searches path suffix `suf`
    and passes along `kwargs`, only in `folders`, if specified."""

    def _inner(o, recurse=recurse, folders=folders):
        return get_audio_files(o / suf, recurse, folders)

    return _inner


URLs.SPEAKERS10 = "http://www.openslr.org/resources/45/ST-AEDS-20180100_1-OS.tgz"
URLs.ESC50 = "https://github.com/karoldvl/ESC-50/archive/master.zip"
URLs.SAMPLE_SPEAKERS10 = (
    "https://github.com/fastaudio/10_Speakers_Sample/archive/10_speakers_sample.zip"
)


def tar_extract_at_filename(fname, dest):
    "Extract `fname` to `dest`/`fname.name` folder using `tarfile`"
    dest = Path(dest) / Path(fname).with_suffix("").name
    tarfile.open(fname, "r:gz").extractall(dest)


# fix to preserve metadata for subclass tensor in serialization
# src: https://github.com/fastai/fastai/pull/3383
# TODO: remove this when #3383 lands and a new fastai version is created
def _rebuild_from_type(func, type, args, dict):
    ret = func(*args).as_subclass(type)
    ret.__dict__ = dict
    return ret


@patch
def __reduce_ex__(self: TensorBase, proto):
    torch.utils.hooks.warn_if_has_hooks(self)
    args = (
        type(self),
        self.storage(),
        self.storage_offset(),
        tuple(self.size()),
        self.stride(),
    )
    if self.is_quantized:
        args = args + (self.q_scale(), self.q_zero_point())
    args = args + (self.requires_grad, OrderedDict())
    f = _fa_rebuild_qtensor if self.is_quantized else _fa_rebuild_tensor
    return (_rebuild_from_type, (f, type(self), args, self.__dict__))


class AudioTensor(TensorBase):
    """
    Semantic torch tensor that represents an audio.
    Contains all of the functionality of a normal tensor,
    but additionally can be created from files and has
    extra properties. Also knows how to show itself.
    """

    @classmethod
    @delegates(torchaudio.load, keep=True)
    def create(cls, fn, cache_folder=None, **kwargs):
        "Creates audio tensor from file"
        if cache_folder is not None:
            fn = cache_folder / fn.name
        sig, sr = torchaudio.load(fn, **kwargs)
        return cls(sig, sr=sr)

    def __new__(cls, x, sr=None, **kwargs):
        return super().__new__(cls, x, sr=sr, **kwargs)

    @property
    def nsamples(self):
        return self.shape[-1]

    @property
    def nchannels(self):
        return self.shape[-2]

    @property
    def duration(self):
        return self.nsamples / float(self.sr)

    def hear(self):
        "Listen to audio clip. Creates a html player."
        display(Audio(self.cpu(), rate=self.sr))

    def show(self, ctx=None, hear=True, **kwargs):
        """Show audio clip using librosa.
        Pass `hear=True` to also display a html
        player to listen.
        """
        if hear:
            self.hear()
        return show_audio_signal(self, ctx=ctx, **kwargs)

    def apply_gain(self, gain):
        self.data *= gain
        return self

    def cutout(self, cut_pct):
        mask = torch.zeros(int(self.nsamples * cut_pct))
        mask_start = random.randint(0, self.nsamples - len(mask))
        self.data[:, mask_start : mask_start + len(mask)] = mask
        return self

    def lose_signal(self, loss_pct):
        mask = (torch.rand_like(self.data[0]) > loss_pct).float()
        self.data[..., :] *= mask
        return self

    def save(self, fn: Path, overwrite=True):
        "Save the audio into the specfied path"
        fn = path.expanduser(fn)
        if not overwrite and path.exists(fn):
            raise Exception("File already exists")
        torchaudio.save(fn, self.data, self.sr)


def show_audio_signal(ai, ctx, ax=None, title="", **kwargs):
    ax = ifnone(ax, ctx)
    if ax is None:
        _, ax = plt.subplots()
    ax.axis(False)
    for i, channel in enumerate(ai):
        # x_start, y_start, x_lenght, y_lenght, all in percent
        ia = ax.inset_axes((i / ai.nchannels, 0.2, 1 / ai.nchannels, 0.7))
        waveplot(channel.cpu().numpy(), ai.sr, ax=ia, **kwargs)
        ia.set_title(f"Channel {i}")
    ax.set_title(title)

    return ax


@typedispatch
def show_batch(
    x: AudioTensor,
    y,
    samples,
    ctxs=None,
    max_n=6,
    nrows=2,
    ncols=1,
    figsize=None,
    **kwargs,
):
    if figsize is None:
        figsize = (4 * x.nchannels, 6)

    if ctxs is None:
        ctxs = get_grid(
            min(len(samples), max_n), nrows=nrows, ncols=ncols, figsize=figsize
        )
    ctxs = show_batch[object](
        x, y, samples, ctxs=ctxs, max_n=max_n, hear=False, **kwargs
    )
    return ctxs


class OpenAudio(Transform):
    """
    Transform that creates AudioTensors from a list of files.
    """

    def __init__(self, items):
        self.items = items

    def encodes(self, i):
        o = self.items[i]
        return AudioTensor.create(o)

    def decodes(self, i) -> Path:
        return self.items[i]



import torch

# Must be imported explicitly to override the top-level `torch.fft` function
import torch.fft
from torch import Tensor


def region_mask(n, min_mask_size, max_mask_size, maskable_length, device=None):
    """Create a vector of ``n`` random region masks.

    Masks are returned in a vector with shape `(n, maskable_length)`.

    The mask vectors are boolean vectors of ``maskable_length`` with a
    continuous region of 1s between ``min_mask_size`` and ``max_mask_size``
    (inclusive).

    """
    # Generate the start & end positions for each mask, then compare these to
    # absolute indices to create the actual mask vectors.
    mask_sizes = (
        torch.rand([n, 1], device=device) * (max_mask_size - min_mask_size)
        + min_mask_size
    )
    mask_starts = torch.rand([n, 1], device=device) * (maskable_length - mask_sizes)
    mask_ends = mask_starts + mask_sizes
    indexes = torch.arange(0, maskable_length, device=device)
    return (mask_starts <= indexes) & (indexes < mask_ends)


def mask_along_axis_(specgrams, num_masks, min_size, max_size, mask_val=None, axis=2):
    """Apply SpecAugment masks emitting from one axis.

    ``specgrams`` should be a tensor of shape (batch, channels, freq, time).
    ``axis`` must be either `2` or `3`.

    Masks are replaced with the mean of the masked area. Provide ``mask_val`` to
    mask with a specific value.

    """
    device = specgrams.device

    if axis != 3:
        # Orient so the axis we're masking comes last
        specgrams = specgrams.transpose(axis, -1)

    n, _, _, a = specgrams.shape

    # First create the broadcastable masks. Each spectrogram gets a different
    # set of masks (but the same masks span across channels).
    if num_masks == 1:
        masks = region_mask(n, min_size, max_size, a, device=device)
    else:
        # To create multiple masks per spectrogram, create a larger batch,
        # reshape, then merge.
        masks = (
            region_mask(
                num_masks * n,
                min_size,
                max_size,
                a,
                device=device,
            )
            .view(num_masks, n, a)
            .amax(dim=0)
        )
    # Expand so it can be broadcasted.
    masks = masks.view(n, 1, 1, a)

    if mask_val:
        specgrams.masked_fill_(masks, mask_val)
    else:
        # Mask with the channel-wise mean. Note while each channel takes the
        # same mask shape, the replacement value is determined per-channel.

        # Take the mean of the masked area, not the whole spectrogram, so as not
        # to change the overall mean (although this will affect the standard
        # deviation).
        mask_vals = (
            specgrams.mul(masks).sum((-2, -1))
            # This will be broadcast, so we have to manually multiply
            # by the size of that dimension.
            / (masks.sum((-2, -1)) * specgrams.shape[-2])
        )

        # Alternate method: mean of whole channel, not just masked area.
        # mask_vals = specgrams.mean(-2, -1)

        specgrams = torch.where(masks, mask_vals[..., None, None], specgrams)
    if axis == 3:
        return specgrams
    else:
        # Restore original orientation
        return specgrams.transpose(axis, -1)


def random_mask(shape, p, device=None):
    """Create a bool tensor with items set to 1 with probability ``p``.

    (Items are set individually.)

    ``p`` may be a float or a tensor broadcastable to ``shape``.

    """
    return torch.rand(shape, device=device) <= p


# TODO: Remove this & all uses, replace with torch.fft.rfftfreq
def _rfftfreq(n, d=1.0, device=None):
    """Get the sample frequencies for a discrete fourier transform."""
    val = 1.0 / (n * d)
    results = torch.arange(0, n // 2 + 1, device=device, dtype=int)
    return results * val


class NoiseColor:
    """Exponent beta values for named noise colors (see `colored_noise`)."""

    Violet = -2
    Blue = -1
    White = 0
    Pink = 1
    Brown = 2

    @staticmethod
    def valid(value: int) -> bool:
        """Is ``value`` a valid color for noise?"""
        return value in [*range(-2, 3)]


def colored_noise(shape, exponent, fmin=0, device=None):
    """Gaussian (1/f)**beta noise.

    Based on the algorithm in:
    Timmer, J. and Koenig, M.:
    On generating power law noise.
    Astron. Astrophys. 300, 707-710 (1995)
    Normalised to unit variance

    Ported to PyTorch from Felix Patzelt's numpy implementation, MIT Licensed.
    https://github.com/felixpatzelt/colorednoise/blob/acc9ec529b181dec2bd7fb914f87f0a62a0553d3/colorednoise.py

    Parameters:
    -----------

    exponent : float
        The power-spectrum of the generated noise is proportional to
        S(f) = (1 / f)**beta
        flicker / pink noise:   exponent beta = 1
        brown noise:            exponent beta = 2
        Furthermore, the autocorrelation decays proportional to lag**-gamma
        with gamma = 1 - beta for 0 < beta < 1.
        There may be finite-size issues for beta close to one.

    shape : int or iterable
        The output has the given shape, and the desired power spectrum in
        the last coordinate. That is, the last dimension is taken as time,
        and all other components are independent.

    fmin : float, optional
        Low-frequency cutoff.
        Default: 0 corresponds to original paper. It is not actually
        zero, but 1/samples.

    Returns
    -------

    out : array
        The samples.

    Examples:
    ---------

    # generate 1/f noise == pink noise == flicker noise
    >>> y = colored_noise([1], 5)

    """
    # Use `is` to allow for tensor exponents
    if exponent is NoiseColor.White:
        # White noise is simple - use a faster method.
        return torch.randn(shape)

    # The number of samples in each time series
    nsamples = shape[-1]

    # Calculate Frequencies (we asume a sample rate of one)
    # Use fft functions for real output (-> hermitian spectrum)
    # FIXME: torch.fft is missing this for some reason
    # f = torch.fft.rfftfreq(nsamples)
    f = _rfftfreq(nsamples, device=device)

    # Build scaling factors for all frequencies
    s_scale = f
    fmin = max(fmin, 1.0 / nsamples)  # Low frequency cutoff
    ix = (s_scale < fmin).sum()  # Index of the cutoff
    if ix and ix < len(s_scale):
        s_scale[:ix] = s_scale[ix]
    s_scale = s_scale ** (-exponent / 2.0)

    # Calculate theoretical output standard deviation from scaling
    w = s_scale[1:].clone()
    w[-1] *= (1 + (nsamples % 2)) / 2.0  # correct f = +-0.5
    sigma = 2 * (w ** 2).sum().sqrt() / nsamples

    # Adjust size to generate one Fourier component per frequency
    new_shape = (*shape[:-1], f.size(0))
    # Original numpy imp
    # shape[-1] = f.size(0)

    # Add empty dimension(s) to broadcast s_scale along last
    # dimension of generated random power + phase (below)
    dims_to_add = len(new_shape) - 1
    s_scale = s_scale[[None] * dims_to_add + [Ellipsis]]

    # Generate scaled random power + phase
    sr = torch.randn(new_shape, device=device) * s_scale
    si = torch.randn(new_shape, device=device) * s_scale

    # If the signal length is even, frequencies +/- 0.5 are equal
    # so the coefficient must be real.
    if not (nsamples % 2):
        si[..., -1] = 0

    # Regardless of signal length, the DC component must be real
    si[..., 0] = 0

    # Combine power + corrected phase to Fourier components
    s = sr + 1j * si

    # Transform to real time series & scale to unit variance
    y = torch.fft.irfft(s, n=nsamples, dim=-1) / sigma

    return y


def add_noise_(x: Tensor, color: NoiseColor, min_level: float, max_level: float, p=1.0):
    """Add colored noise to a batch of tensors, randomising per-item.

    Levels are relative to the standard deviation of each item.

    """
    n = x.size(0)

    noise_levels = (
        # Note that noise is scaled per-item, not per-channel. This means loud
        # channels will dwarf quiet channels.
        random_mask([n], p, device=x.device)
        * x.reshape(n, -1).std(dim=1)
        * (torch.rand([n], device=x.device) * (max_level - min_level) + min_level)
    ).reshape([n] + [1] * (x.dim() - 1))
    x += noise_levels * (colored_noise(x.shape, exponent=color, device=x.device) - 0.5)
    return x



import torchaudio
import warnings
from dataclasses import asdict, is_dataclass
from fastai.data.core import TensorImageBase
from fastai.imports import inspect, partial, plt
from fastai.vision.data import get_grid
from fastcore.dispatch import typedispatch
from fastcore.meta import delegates
from fastcore.transform import Transform
from fastcore.utils import ifnone
from inspect import signature
from librosa.display import specshow
from torch import nn

from .signal import AudioTensor


class AudioSpectrogram(TensorImageBase):
    """
    Semantic torch tensor that represents an Audio Spectrogram.
    Contains all of the functionality of a normal tensor,
    but has extra properties and knows how to show itself.
    """

    @classmethod
    def create(cls, sg_tensor, settings=None):
        """Create an AudioSpectrogram from a torch tensor"""
        audio_sg = cls(sg_tensor)
        audio_sg._settings = settings
        return audio_sg

    @property
    def duration(self):
        # spectrograms round up length to fill incomplete columns,
        # so we subtract 0.5 to compensate, wont be exact
        return (self.hop_length * (self.shape[-1] - 0.5)) / self.sr

    @property
    def width(self):
        return self.shape[-1]

    @property
    def height(self):
        return self.shape[-2]

    @property
    def nchannels(self):
        return self.shape[-3]

    def _all_show_args(self, show_y: bool = True):
        proper_kwargs = get_usable_kwargs(
            specshow, self._settings, exclude=["ax", "kwargs", "data"]
        )
        if "mel" not in self._settings or not show_y:
            y_axis = None
        else:
            y_axis = "mel" if self.mel else "linear"
        proper_kwargs.update({"x_axis": "time", "y_axis": y_axis})
        proper_kwargs.update(self._show_args)
        return proper_kwargs

    @property
    def _colorbar_fmt(self):
        return "%+2.0f dB" if "to_db" in self._settings and self.to_db else "%+2.0f"

    def __getattr__(self, name):
        if name == "settings":
            return self._settings
        if not name.startswith("_"):
            return self._settings[name]
        raise AttributeError(
            f"{self.__class__.__name__} object has no attribute {name}"
        )

    def show(self, ctx=None, ax=None, title="", **kwargs):
        "Show spectrogram using librosa"
        return show_spectrogram(self, ctx=ctx, ax=ax, title=title, **kwargs)


def show_spectrogram(sg, title="", ax=None, ctx=None, **kwargs):
    ax = ifnone(ax, ctx)
    if ax is None:
        _, ax = plt.subplots()
    ax.axis(False)
    for i, channel in enumerate(sg):
        # x_start, y_start, x_lenght, y_lenght, all in percent
        ia = ax.inset_axes((i / sg.nchannels, 0.2, 1 / sg.nchannels, 0.7))
        z = specshow(
            channel.cpu().numpy(), ax=ia, **sg._all_show_args(show_y=i == 0), **kwargs
        )
        ia.set_title(f"Channel {i}")
        if i == 0:  # Only colorbar the first one
            plt.colorbar(z, format=sg._colorbar_fmt, ax=ax)
    ax.set_title(title)

    return ax


@typedispatch
def show_batch(
    x: AudioSpectrogram,
    y,
    samples,
    ctxs=None,
    max_n=6,
    nrows=2,
    ncols=1,
    figsize=None,
    **kwargs,
):
    if figsize is None:
        figsize = (4 * x.nchannels, 4)

    if ctxs is None:
        ctxs = get_grid(
            min(len(samples), max_n), nrows=nrows, ncols=ncols, figsize=figsize
        )
    ctxs = show_batch[object](x, y, samples, ctxs=ctxs, max_n=max_n, **kwargs)
    return ctxs


_GenSpec = torchaudio.transforms.Spectrogram
_GenMelSpec = torchaudio.transforms.MelSpectrogram
_GenMFCC = torchaudio.transforms.MFCC
_ToDB = torchaudio.transforms.AmplitudeToDB


class AudioToSpec(Transform):
    """
    Transform to create spectrograms from audio tensors.
    """

    def __init__(self, pipe, settings):
        self.pipe = pipe
        self.settings = settings

    @classmethod
    def from_cfg(cls, audio_cfg):
        "Creates AudioToSpec from configuration file"
        cfg = asdict(audio_cfg) if is_dataclass(audio_cfg) else dict(audio_cfg)
        transformer = SpectrogramTransformer(mel=cfg.pop("mel"), to_db=cfg.pop("to_db"))
        return transformer(**cfg)

    def encodes(self, audio: AudioTensor):
        self.pipe.to(audio.device)
        self.settings.update({"sr": audio.sr, "nchannels": audio.nchannels})
        return AudioSpectrogram.create(self.pipe(audio), settings=dict(self.settings))


def SpectrogramTransformer(mel=True, to_db=True):
    """Creates a factory for creating AudioToSpec
    transforms with different parameters"""
    sg_type = {"mel": mel, "to_db": to_db}
    transforms = _get_transform_list(sg_type)
    pipe_noargs = partial(fill_pipeline, sg_type=sg_type, transform_list=transforms)
    pipe_noargs.__signature__ = _get_signature(transforms)
    return pipe_noargs


def _get_transform_list(sg_type):
    """Builds a list of higher-order transforms with no arguments"""
    transforms = []
    if sg_type["mel"]:
        transforms.append(_GenMelSpec)
    else:
        transforms.append(_GenSpec)
    if sg_type["to_db"]:
        transforms.append(_ToDB)
    return transforms


def fill_pipeline(transform_list, sg_type, **kwargs):
    """Adds correct args to each transform"""
    kwargs = _override_bad_defaults(dict(kwargs))
    function_list = []
    settings = {}
    for f in transform_list:
        usable_kwargs = get_usable_kwargs(f, kwargs)
        function_list.append(f(**usable_kwargs))
        settings.update(usable_kwargs)
    warn_unused(kwargs, settings)
    return AudioToSpec(nn.Sequential(*function_list), settings={**sg_type, **settings})


def _get_signature(transforms):
    """Looks at transform list and extracts all valid args for tab completion"""
    delegations = [delegates(to=f, keep=True) for f in transforms]
    out = lambda **kwargs: None  # noqa: E731
    for d in delegations:
        out = d(out)
    return signature(out)


def _override_bad_defaults(kwargs):
    if "n_fft" not in kwargs or kwargs["n_fft"] is None:
        kwargs["n_fft"] = 1024
    if "win_length" not in kwargs or kwargs["win_length"] is None:
        kwargs["win_length"] = kwargs["n_fft"]
    if "hop_length" not in kwargs or kwargs["hop_length"] is None:
        kwargs["hop_length"] = int(kwargs["win_length"] / 2)
    return kwargs


def warn_unused(all_kwargs, used_kwargs):
    unused_kwargs = set(all_kwargs.keys()) - set(used_kwargs.keys())
    for kwarg in unused_kwargs:
        warnings.warn(f"{kwarg} is not a valid arg name and was not used")


def get_usable_kwargs(func, kwargs, exclude=None):
    exclude = ifnone(exclude, [])
    defaults = {
        k: v.default
        for k, v in inspect.signature(func).parameters.items()
        if k not in exclude
    }
    usable = {k: v for k, v in kwargs.items() if k in defaults}
    return {**defaults, **usable}


@delegates(_GenMFCC.__init__)
class AudioToMFCC(Transform):
    """
    Transform to create MFCC features from audio tensors.
    """

    def __init__(self, **kwargs):
        func_args = get_usable_kwargs(_GenMFCC, kwargs, [])
        self.transformer = _GenMFCC(**func_args)
        self.settings = func_args

    @classmethod
    def from_cfg(cls, audio_cfg):
        "Creates AudioToMFCC from configuration file"
        cfg = asdict(audio_cfg) if is_dataclass(audio_cfg) else audio_cfg
        return cls(**cfg)

    def encodes(self, x: AudioTensor):
        sg_settings = {"sr": x.sr, "nchannels": x.nchannels, **self.settings}
        return AudioSpectrogram.create(
            self.transformer(x).detach(), settings=sg_settings
        )



import librosa
import torch
import warnings
from fastai.imports import partial, random
from fastcore.transform import Transform
from fastcore.utils import ifnone
from torch.nn import functional as F
from torchaudio.functional import compute_deltas

from ..core.spectrogram import AudioSpectrogram, AudioTensor
from ..util import auto_batch
from .functional import mask_along_axis_
from .signal import AudioPadType


class SpectrogramTransform(Transform):
    "Helps prevent us trying to apply to Audio Tensors"

    def encodes(self, audio: AudioTensor) -> AudioTensor:
        warnings.warn(
            f"You are trying to apply a {type(self).__name__} to an AudioTensor {audio.shape}"
        )
        return audio


class CropTime(SpectrogramTransform):
    """Random crops full spectrogram to be length specified in ms by crop_duration"""

    def __init__(self, duration, pad_mode=AudioPadType.Zeros):
        self.duration = duration
        self.pad_mode = pad_mode

    def encodes(self, sg: AudioSpectrogram) -> AudioSpectrogram:
        sr, hop = sg.sr, sg.hop_length
        w_crop = int((sr * self.duration) / (1000 * hop)) + 1
        w_sg = sg.shape[-1]
        if w_sg == w_crop:
            sg_crop = sg
        elif w_sg < w_crop:
            sg_crop = _tfm_pad_spectro(sg, w_crop, pad_mode=self.pad_mode)
        else:
            crop_start = random.randint(0, int(w_sg - w_crop))
            sg_crop = sg[:, :, crop_start : crop_start + w_crop]
            sg_crop.sample_start = int(crop_start * hop)
            sg_crop.sample_end = sg_crop.sample_start + int(self.duration * sr)
        sg.data = sg_crop
        return sg


def _tfm_pad_spectro(sg, width, pad_mode=AudioPadType.Zeros):
    """Pad spectrogram to specified width, using specified pad mode"""
    c, y, x = sg.shape
    if pad_mode in [AudioPadType.Zeros, AudioPadType.Zeros_After]:
        padded = torch.zeros((c, y, width))
        start = random.randint(0, width - x) if pad_mode == AudioPadType.Zeros else 0
        padded[:, :, start : start + x] = sg.data
        return padded
    elif pad_mode == AudioPadType.Repeat:
        repeats = width // x + 1
        return sg.repeat(1, 1, repeats)[:, :, :width]
    else:
        raise ValueError(
            f"""pad_mode {pad_mode} not currently supported,
            only AudioPadType.Zeros, AudioPadType.Zeros_After,
            or AudioPadType.Repeat"""
        )


class MaskFreq(SpectrogramTransform):
    """Google SpecAugment frequency masking from https://arxiv.org/abs/1904.08779."""

    def __init__(self, num_masks=1, size=20, start=None, val=None):
        self.num_masks = num_masks
        self.size = size
        self.start = start
        self.val = val

    def encodes(self, sg: AudioSpectrogram) -> AudioSpectrogram:
        channel_mean = sg.contiguous().view(sg.size(0), -1).mean(-1)[:, None, None]
        mask_val = ifnone(self.val, channel_mean)
        c, y, x = sg.shape
        # Position of the first mask
        start = ifnone(self.start, random.randint(0, y - self.size))
        for _ in range(self.num_masks):
            mask = torch.ones(self.size, x) * mask_val
            if not 0 <= start <= y - self.size:
                raise ValueError(
                    f"Start value '{start}' out of range for AudioSpectrogram of shape {sg.shape}"
                )
            sg[:, start : start + self.size, :] = mask
            # Setting start position for next mask
            start = random.randint(0, y - self.size)
        return sg


class MaskTime(SpectrogramTransform):
    """Google SpecAugment time masking from https://arxiv.org/abs/1904.08779."""

    def __init__(self, num_masks=1, size=20, start=None, val=None):
        self.num_masks = num_masks
        self.size = size
        self.start = start
        self.val = val

    def encodes(self, sg: AudioSpectrogram) -> AudioSpectrogram:
        sg.data = torch.einsum("...ij->...ji", sg)
        sg.data = MaskFreq(
            self.num_masks,
            self.size,
            self.start,
            self.val,
        )(sg)
        sg.data = torch.einsum("...ij->...ji", sg)
        return sg


class _MaskAxisGPU(SpectrogramTransform):
    """Base class for GPU-based SpecAugment masking transforms."""

    def __init__(self, axis, num_masks, min_size, max_size, mask_val):
        if axis not in [2, 3]:
            raise ValueError("Can only mask the time or frequency axis (2 or 3)")
        self.num_masks = num_masks
        self.min_size = min_size
        self.max_size = max_size
        self.mask_val = mask_val
        self.axis = axis
        super().__init__()

    @auto_batch(3)
    def encodes(self, sg: AudioSpectrogram):
        return mask_along_axis_(
            sg,
            num_masks=self.num_masks,
            min_size=self.min_size,
            max_size=self.max_size,
            mask_val=self.mask_val,
            axis=self.axis,
        )


class MaskFreqGPU(_MaskAxisGPU):
    """Google SpecAugment frequency masking from https://arxiv.org/abs/1904.08779.

    This version runs on batches and can be run efficiently on the GPU.

    """

    def __init__(self, num_masks=1, min_size=1, max_size=10, mask_val=None):
        super().__init__(2, num_masks, min_size, max_size, mask_val)


class MaskTimeGPU(_MaskAxisGPU):
    """Google SpecAugment time masking from https://arxiv.org/abs/1904.08779.

    This version runs on batches and can be run efficiently on the GPU.

    """

    def __init__(self, num_masks=1, min_size=1, max_size=10, mask_val=None):
        super().__init__(3, num_masks, min_size, max_size, mask_val)


class SGRoll(SpectrogramTransform):
    """Shifts spectrogram along x-axis wrapping around to other side"""

    def __init__(self, max_shift_pct=0.5, direction=0):
        if int(direction) not in [-1, 0, 1]:
            raise ValueError("Direction must be -1(left) 0(bidirectional) or 1(right)")
        self.max_shift_pct = max_shift_pct
        self.direction = direction

    def encodes(self, sg: AudioSpectrogram) -> AudioSpectrogram:
        direction = random.choice([-1, 1]) if self.direction == 0 else self.direction
        w = sg.shape[-1]
        roll_by = int(w * random.random() * self.max_shift_pct * direction)
        sg.data = sg.roll(roll_by, dims=-1)
        return sg


def _torchdelta(sg: AudioSpectrogram, order=1, width=9):
    """Converts to numpy, takes delta and converts back to torch, needs torchification"""
    if sg.shape[1] < width:
        raise ValueError(
            f"""Delta not possible with current settings, inputs must be wider than
        {width} columns, try setting max_to_pad to a larger value to ensure a minimum width"""
        )
    return AudioSpectrogram(
        torch.from_numpy(librosa.feature.delta(sg.numpy(), order=order, width=width))
    )


class Delta(SpectrogramTransform):
    """Creates delta with order 1 and 2 from spectrogram
    and concatenate with the original"""

    def __init__(self, width=9):
        self.td = partial(_torchdelta, width=width)

    def encodes(self, sg: AudioSpectrogram):
        new_channels = [
            torch.stack([c, self.td(c, order=1), self.td(c, order=2)]) for c in sg
        ]
        sg.data = torch.cat(new_channels, dim=0)
        return sg


class DeltaGPU(SpectrogramTransform):
    """Adds extra channels with delta (orders 1 and 2) to spectrogram."""

    def __init__(self, width=9, mode="replicate"):
        self.width = width
        self.mode = mode

    @auto_batch(3)
    def encodes(self, sg: AudioSpectrogram):
        delta = compute_deltas(sg, win_length=self.width, mode=self.mode)
        delta2 = compute_deltas(delta, win_length=self.width, mode=self.mode)
        sg.data = torch.cat([sg, delta, delta2], dim=1).contiguous()
        return sg


class TfmResize(SpectrogramTransform):
    """Temporary fix to allow image resizing transform"""

    def __init__(self, size, interp_mode="bilinear"):
        self.size = size
        self.interp_mode = interp_mode

    def encodes(self, sg: AudioSpectrogram) -> AudioSpectrogram:
        if isinstance(self.size, int):
            self.size = (self.size, self.size)
        c, y, x = sg.shape
        sg.data = F.interpolate(
            sg.unsqueeze(0), size=self.size, mode=self.interp_mode, align_corners=False
        ).squeeze(0)
        return sg


class TfmResizeGPU(SpectrogramTransform):
    """Resize the spectrogram using interpolation."""

    def __init__(self, size, interp_mode="bilinear"):
        if isinstance(size, int):
            self.size = (size, size)
        else:
            self.size = size
        self.interp_mode = interp_mode

    @auto_batch(3)
    def encodes(self, sg: AudioSpectrogram):
        sg.data = F.interpolate(
            sg, size=self.size, mode=self.interp_mode, align_corners=False
        )
        return sg



import os

from fastaudio.all import (
    AudioConfig,
    AudioToSpec,
    OpenAudio,
    Pipeline,
    audio_item_tfms,
    preprocess_audio_folder
)
from fastaudio.util import test_audio_tensor


def test_basic_config():
    "Make sure mel setting is passed down and is false for normal spectro"
    sg_cfg = AudioConfig.BasicSpectrogram()
    assert sg_cfg.mel == False


def test_load_audio_with_basic_config():
    """
    Grab a random file, test that the n_fft are passed successfully
    via config and stored in sg settings
    """
    sg_cfg = AudioConfig.BasicSpectrogram(n_fft=2000, hop_length=155)
    a2sg = AudioToSpec.from_cfg(sg_cfg)
    audio = test_audio_tensor()
    sg = a2sg(audio)
    assert sg.n_fft == sg_cfg.n_fft
    assert sg.width == int(audio.nsamples / sg_cfg.hop_length) + 1


def test_basic_pipeline():
    cfg = {"mel": False, "to_db": False, "hop_length": 128, "n_fft": 400}
    test_audio_tensor().save("./test.wav")
    f = "./test.wav"
    oa = OpenAudio([f])
    a2s = AudioToSpec.from_cfg(cfg)
    db_mel_pipe = Pipeline([oa, a2s])
    assert db_mel_pipe(0).hop_length == cfg["hop_length"]


def test_basic_pre_audio():
    tfms = audio_item_tfms(8000, True, 4000)
    assert len(tfms) == 3


def test_pre_process_audio():
    d = "data_test"
    if not os.path.isdir(d):
        os.mkdir(d)
    test_audio_tensor().save(d + "/test.wav")
    preprocess_audio_folder(d)



import torch
from fastai.data.all import test_eq as _test_eq
from unittest.mock import patch

from fastaudio.augment.functional import region_mask


class TestCreateRegionMask:
    def test_shape(self):
        _test_eq(region_mask(1, 5, 7, 10).shape, (1, 10))
        _test_eq(region_mask(2, 3, 7, 12).shape, (2, 12))
        _test_eq(region_mask(4, 0, 3, 3).shape, (4, 3))

    def test_max(self):
        # Test max size
        with patch(
            "torch.rand",
            side_effect=[
                torch.Tensor([[[[1.0]]]]),
                torch.Tensor([[[[0.0]]]]),
            ],
        ):
            _test_eq(
                region_mask(1, 4, 6, 10),
                torch.BoolTensor([[[[1] * 6 + [0] * 4]]]),
            )

    def test_min(self):
        # Test min size
        with patch(
            "torch.rand",
            side_effect=[
                torch.Tensor([0.0]),
                # Test start middle start here too
                torch.Tensor([0.5]),
            ],
        ):
            _test_eq(
                region_mask(1, 4, 6, 10),
                torch.BoolTensor([0] * 3 + [1] * 4 + [0] * 3),
            )

    def test_multiple_masks(self):
        # Test multiple masks
        with patch(
            "torch.rand",
            side_effect=[
                torch.Tensor([[1.0], [0.0]]),
                torch.Tensor([[0.0], [0.5]]),
            ],
        ):
            _test_eq(
                region_mask(2, 4, 6, 10),
                torch.BoolTensor([[1] * 6 + [0] * 4, [0] * 3 + [1] * 4 + [0] * 3]),
            )



import pytest

import random
import torch
from fastai.data.all import test_close as _test_close
from fastai.data.all import test_eq as _test_eq
from fastai.data.all import test_fail as _test_fail
from fastai.data.all import test_ne as _test_ne
from unittest.mock import patch

from fastaudio.all import (
    AudioConfig,
    AudioPadType,
    AudioSpectrogram,
    AudioToSpec,
    CropTime,
    DeltaGPU,
    MaskFreqGPU,
    MaskTimeGPU,
    OpenAudio,
    Pipeline,
    ResizeSignal,
    SGRoll,
    SignalShifter,
    SpectrogramTransformer,
    TfmResizeGPU,
)
from fastaudio.util import apply_transform, test_audio_tensor


def test_crop_time():
    for i in [1, 2, 5]:
        a2s = AudioToSpec.from_cfg(AudioConfig.Voice())
        audio = test_audio_tensor(seconds=3)
        crop = CropTime(i * 1000)
        inp, out = apply_transform(crop, a2s(audio))
        _test_eq(i, round(out.duration))
        _test_close(out.width, int((i / inp.duration) * inp.width), eps=1.01)


def test_crop_time_with_pipeline():
    """
    AudioToSpec->CropTime and ResizeSignal->AudioToSpec
    will result in same size images
    """
    afn = "./test.wav"
    test_audio_tensor().save(afn)
    ex_files = [afn] * 4
    oa = OpenAudio(ex_files)
    crop_dur = random.randint(1000, 5000)
    DBMelSpec = SpectrogramTransformer(mel=True, to_db=True)
    pipe_cropsig = Pipeline([oa, DBMelSpec(hop_length=128), CropTime(crop_dur)])
    pipe_cropspec = Pipeline(
        [
            oa,
            ResizeSignal(crop_dur),
            DBMelSpec(hop_length=128),
        ]
    )
    for i in range(4):
        _test_eq(pipe_cropsig(i).width, pipe_cropspec(i).width)


def test_crop_time_after_padding():
    sg_orig = test_audio_tensor()
    a2s = AudioToSpec.from_cfg(AudioConfig.Voice())
    sg = a2s(sg_orig)
    crop_time = CropTime((sg.duration + 5) * 1000, pad_mode=AudioPadType.Zeros_After)
    inp, out = apply_transform(crop_time, sg.clone())
    _test_ne(sg.duration, sg_orig.duration)


def test_crop_time_repeat_padding():
    "Test that repeat padding works when cropping time"
    repeat = 3
    audio = test_audio_tensor()
    crop_12000ms_repeat = CropTime(
        repeat * 1000 * audio.duration, pad_mode=AudioPadType.Repeat
    )
    a2s = AudioToSpec.from_cfg(AudioConfig.Voice())
    sg = a2s(audio)
    inp, out = apply_transform(crop_12000ms_repeat, sg)
    _test_eq(inp.width, sg.width)
    _test_ne(sg.width, out.width)


def test_fail_bad_pad():
    # test bad pad_mode doesnt fail silently, correct is 'zeros_after'
    _test_fail(CropTime(12000, pad_mode="zerosafter"))


def test_mask_freq():
    c, f, t = 2, 120, 80

    # Create a random frequency mask and test that it is being correctly applied
    min_size = 5
    max_size = 7

    sg = AudioSpectrogram(torch.rand([c, f, t]))
    val = 10  # Use a value not in the original spectrogram
    gradient_sg = AudioSpectrogram(
        torch.linspace(0, 1, f).view(1, f, 1).repeat([c, 1, t])
    )
    ones = torch.ones_like(sg)

    # Test patching with mean
    with patch(
        "fastaudio.functional.region_mask",
        side_effect=[
            torch.BoolTensor([[1] * 10 + [0] * (f - 10)]),
        ],
    ):
        mask_with_mean = MaskFreqGPU(
            min_size=min_size, max_size=max_size, mask_val=None
        )
        # Use a gradient so we can be sure the mean will never show up outside the mask
        inp, out = apply_transform(mask_with_mean, gradient_sg)
        channelwise_mean = inp[:, :10, :].mean(dim=(-2, -1)).reshape(-1, 1, 1)
        _test_eq(out[:, :10, :], channelwise_mean * ones[:, :10, :])
        assert not (out[:, 10:, :] == channelwise_mean).any(), out == channelwise_mean

    # Test multiple masks (and patching with value)
    with patch(
        "fastaudio.functional.region_mask",
        side_effect=[
            torch.BoolTensor([[1] * 10 + [0] * (f - 10), [0] * (f - 10) + [1] * 10]),
        ],
    ):
        mask_with_val = MaskFreqGPU(
            min_size=min_size, num_masks=2, max_size=max_size, mask_val=val
        )
        inp, out = apply_transform(mask_with_val, sg)
        _test_eq(
            out[:, :10, :],
            val * ones[:, :10, :],
        )
        _test_eq(
            out[:, f - 10 :, :],
            val * ones[:, f - 10 :, :],
        )
        matches = out[:, 10 : f - 10] == val
        assert not matches.any(), matches


def test_mask_time():
    c, f, t = 2, 120, 80

    min_size = 5
    max_size = 7

    sg = AudioSpectrogram(torch.rand([c, f, t]))
    val = 10  # Use a value not in the original spectrogram
    gradient_sg = AudioSpectrogram(
        torch.linspace(0, 1, t).view(1, 1, t).repeat([c, f, 1])
    )
    ones = torch.ones_like(sg)

    # Test patching with mean
    with patch(
        "fastaudio.functional.region_mask",
        side_effect=[
            torch.BoolTensor([[1] * 10 + [0] * (t - 10)]),
        ],
    ):
        mask_with_mean = MaskTimeGPU(
            min_size=min_size, max_size=max_size, mask_val=None
        )
        # Use a gradient so we can be sure the mean will never show up outside the mask
        inp, out = apply_transform(mask_with_mean, gradient_sg)
        channelwise_mean = inp[..., :10].mean(dim=(-2, -1)).reshape(-1, 1, 1)
        _test_close(
            out[..., :10],
            channelwise_mean * ones[..., :10],
        )
        assert not (out[..., 10:] == channelwise_mean).any(), out == channelwise_mean

    # Test multiple masks (and patching with value)
    with patch(
        "fastaudio.functional.region_mask",
        side_effect=[
            torch.BoolTensor([[1] * 10 + [0] * (t - 10), [0] * (t - 10) + [1] * 10]),
        ],
    ):
        mask_with_val = MaskTimeGPU(
            min_size=min_size, num_masks=2, max_size=max_size, mask_val=val
        )
        inp, out = apply_transform(mask_with_val, sg)
        _test_eq(
            out[..., :10],
            val * ones[..., :10],
        )
        _test_eq(
            out[..., t - 10 :],
            val * ones[..., t - 10 :],
        )
        matches = out[..., 10 : t - 10] == val
        assert not matches.any(), matches


def test_resize_int():
    # Test when size is an int
    size = 224
    resize_int = TfmResizeGPU(size)
    audio = test_audio_tensor()
    a2s = AudioToSpec.from_cfg(AudioConfig.Voice())
    sg = a2s(audio)
    inp, out = apply_transform(resize_int, sg)
    _test_eq(out.shape[1:], torch.Size([size, size]))


def test_delta_channels():
    " nchannels for a spectrogram is how many channels its original audio had "
    delta = DeltaGPU()
    # Explicitly check more than one channel
    audio = test_audio_tensor(channels=2)
    a2s = AudioToSpec.from_cfg(AudioConfig.Voice())
    sg = a2s(audio)
    inp, out = apply_transform(delta, sg)

    _test_eq(out.nchannels, inp.nchannels * 3)
    _test_eq(out.shape[-2:], inp.shape[-2:])
    for i1, i2 in [(0, 2), (1, 3), (0, 4), (1, 5), (2, 4), (3, 5)]:
        assert not torch.allclose(out[i1], out[i2])


def test_signal_shift_on_sg():
    audio = test_audio_tensor()
    a2s = AudioToSpec.from_cfg(AudioConfig.BasicSpectrogram())
    shifter = SignalShifter(1, 1)
    inp, out = apply_transform(shifter, a2s(audio))
    _test_ne(inp, out)


def test_sg_roll():
    roll = SGRoll(1)
    audio = test_audio_tensor()
    a2s = AudioToSpec.from_cfg(AudioConfig.BasicSpectrogram())
    inp, out = apply_transform(roll, a2s(audio))
    _test_ne(inp, out)


def test_sg_roll_fails_direction():
    with pytest.raises(ValueError):
        SGRoll(direction=2)



import random

import pytest

import torch
from fastai.data.all import test_close as _test_close
from fastai.data.all import test_eq as _test_eq

from fastaudio.all import (
    AudioTensor,
    RandTransform,
    RemoveSilence,
    RemoveType,
    Resample,
    ResizeSignal
)
from fastaudio.util import test_audio_tensor


@pytest.fixture(scope="session")
def audio():
    """
    Create a test tensor to be played with. This tensor will only be created
    once at the begging of the tests in this file
    """
    return test_audio_tensor()


def test_path(audio):
    "Check that the audio tensor exists"
    if audio is None:
        raise Exception("Could not generate audio")


def apply_transform(transform, inp):
    """Generate a new input, apply transform, and display/return both input and output"""
    inp_orig = inp.clone()
    out = (
        transform(inp, split_idx=0)
        if isinstance(transform, RandTransform)
        else transform(inp)
    )
    return inp_orig, out


def test_silence_removed(audio):
    "Add silence to a signal and test that it gets removed"
    silencer = RemoveSilence(threshold=20, pad_ms=20)
    orig, silenced = apply_transform(silencer, audio)
    assert silenced.nsamples <= orig.nsamples


def test_silence_not_removed(audio):
    "Test that nothing is removed from audio that doesnt contain silence"
    test_aud = AudioTensor(torch.ones_like(audio), 16000)
    orig_samples = test_aud.nsamples

    for rm_type in [RemoveType.All, RemoveType.Trim, RemoveType.Split]:
        silence_audio_trim = RemoveSilence(rm_type, threshold=20, pad_ms=20)(test_aud)
        assert orig_samples == silence_audio_trim.nsamples


def test_resample(audio):
    no_resample_needed = Resample(audio.sr)
    inp, out = apply_transform(no_resample_needed, audio)
    assert inp.sr == out.sr
    _test_eq(inp.data, out.data)


def test_resample_rates(audio):
    "Test and hear realistic sample rates"
    for rate in [2000, 4000, 8000, 22050, 44100]:
        resampler = Resample(rate)
        inp, out = apply_transform(resampler, audio)
        assert rate == out.sr
        assert out.nsamples == inp.duration * rate


def test_resample_multi_channel(audio):
    audio = test_audio_tensor(channels=3)
    resampler = Resample(8000)
    _, out = apply_transform(resampler, audio)
    _test_eq(out.nsamples, out.duration * 8000)
    _test_eq(out.nchannels, 3)
    _test_eq(out.sr, 8000)


def test_upsample(audio):
    """
    Make sure that the Upsampling is possible. This can
    take a while depending on the target sample rate
    """
    for _ in range(10):
        random_sr = random.randint(16000, 72000)
        random_upsample = Resample(random_sr)(audio)
        num_samples = random_upsample.nsamples
        _test_close(num_samples, abs(audio.nsamples // (audio.sr / random_sr)), eps=1.1)


def test_cropping():
    "Can use the ResizeSignal Transform"
    audio = test_audio_tensor(seconds=10, sr=1000)

    for i in [1, 2, 5]:
        inp, out = apply_transform(ResizeSignal(i * 1000), audio.clone())

        _test_eq(out.duration, i)
        _test_eq(out.nsamples, out.duration * inp.sr)

        # Multi Channel Cropping
        inp, mc = apply_transform(ResizeSignal(i * 1000), audio.clone())
        _test_eq(mc.duration, i)



# -*- coding: utf-8 -*-
from pkg_resources import DistributionNotFound, get_distribution

import os
import torchaudio

torchaudio.USE_SOUNDFILE_LEGACY_INTERFACE = False
# soundfile is torchaudio backend for windows, all other os use sox_io
backend = "soundfile" if os.name == "nt" else "sox_io"
torchaudio.set_audio_backend(backend)
try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = get_distribution(dist_name).version
except DistributionNotFound:
    __version__ = "unknown"
finally:
    del get_distribution, DistributionNotFound



import pytest

import random
import torch
from fastai.data.all import test_close as _test_close
from fastai.data.all import test_eq as _test_eq
from fastai.data.all import test_ne as _test_ne

from fastaudio.all import (
    AddNoiseGPU,
    AudioPadType,
    AudioTensor,
    ChangeVolumeGPU,
    DownmixMono,
    NoiseColor,
    RemoveSilence,
    RemoveType,
    Resample,
    ResizeSignal,
    SignalCutoutGPU,
    SignalLossGPU,
    SignalShifter,
)
from fastaudio.augment.signal import _shift
from fastaudio.util import apply_transform, test_audio_tensor


@pytest.fixture(scope="session")
def audio():
    return test_audio_tensor()


def test_path(audio):
    if audio is None:
        raise Exception("Could not find audio")


def test_silence_removed(audio):
    "Add silence to a signal and test that it gets removed"
    silencer = RemoveSilence(threshold=20, pad_ms=20)
    orig, silenced = apply_transform(silencer, audio)
    assert silenced.nsamples <= orig.nsamples


def test_silence_not_removed(audio):
    "Test that nothing is removed from audio that doesnt contain silence"
    test_aud = AudioTensor(torch.ones_like(audio), 16000)
    orig_samples = test_aud.nsamples

    for rm_type in [RemoveType.All, RemoveType.Trim, RemoveType.Split]:
        silence_audio_trim = RemoveSilence(rm_type, threshold=20, pad_ms=20)(test_aud)
        assert orig_samples == silence_audio_trim.nsamples


def test_resample(audio):
    no_resample_needed = Resample(audio.sr)
    inp, out = apply_transform(no_resample_needed, audio)
    assert inp.sr == out.sr
    _test_eq(inp.data, out.data)


def test_resample_rates(audio):
    "Test and hear realistic sample rates"
    for rate in [2000, 4000, 8000, 22050, 44100]:
        resampler = Resample(rate)
        inp, out = apply_transform(resampler, audio)
        assert rate == out.sr
        assert out.nsamples == inp.duration * rate


def test_resample_multi_channel(audio):
    audio = test_audio_tensor(channels=3)
    resampler = Resample(8000)
    _, out = apply_transform(resampler, audio)
    _test_eq(out.nsamples, out.duration * 8000)
    _test_eq(out.nchannels, 3)
    _test_eq(out.sr, 8000)


def test_upsample(audio):
    for _ in range(10):
        random_sr = random.randint(16000, 72000)
        random_upsample = Resample(random_sr)(audio)
        num_samples = random_upsample.nsamples
        _test_close(num_samples, abs(audio.nsamples // (audio.sr / random_sr)), eps=1.1)


def test_resizing_signal():
    "Can use the ResizeSignal Transform"
    audio = test_audio_tensor(seconds=10, sr=1000)
    mcaudio = test_audio_tensor(channels=2)

    for i in [1, 2, 5]:
        inp, out = apply_transform(ResizeSignal(i * 1000), audio)
        _test_eq(out.duration, i)
        _test_eq(out.nsamples, out.duration * inp.sr)

        inp, out = apply_transform(ResizeSignal(i * 1000), mcaudio)
        _test_eq(out.duration, i)


def test_padding_after_resize(audio):
    "Padding is added to the end  but not the beginning"
    new_duration = (audio.duration + 1) * 1000
    cropsig_pad_after = ResizeSignal(new_duration, pad_mode=AudioPadType.Zeros_After)
    # generate a random input signal that is 3s long
    inp, out = apply_transform(cropsig_pad_after, audio)
    # test end of signal is padded with zeros
    _test_eq(out[:, -10:], torch.zeros_like(out)[:, -10:])
    # test front of signal is not padded with zeros
    _test_ne(out[:, 0:10], out[:, -10:])


def test_padding_both_side_resize(audio):
    "Make sure they are padding on both sides"
    new_duration = (audio.duration + 1) * 1000
    cropsig_pad_after = ResizeSignal(new_duration)
    inp, out = apply_transform(cropsig_pad_after, audio)
    _test_eq(out[:, 0:2], out[:, -2:])


def test_resize_same_duration(audio):
    "Asking to resize to the duration should return the audio back"
    resize = ResizeSignal(audio.duration * 1000)
    inp, out = apply_transform(resize, audio)
    _test_eq(inp, out)


def test_resize_signal_repeat(audio):
    """
    Test pad_mode repeat by making sure that columns are
    equal at the appropriate offsets
    """
    dur = audio.duration * 1000
    repeat = 3
    cropsig_repeat = ResizeSignal(dur * repeat, pad_mode=AudioPadType.Repeat)
    inp, out = apply_transform(cropsig_repeat, audio)
    for i in range(repeat):
        s = int(i * inp.nsamples)
        e = int(s + inp.nsamples)
        _test_eq(out[:, s:e], inp)


def test_fail_invalid_pad_mode():
    with pytest.raises(ValueError):
        ResizeSignal(12000, pad_mode="tenchify")


def test_shift():
    t1 = torch.tensor([[1, 2, 3, 4, 5, 6, 7, 8, 9, 10]])
    t3 = torch.tensor(
        [
            [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            [11, 12, 13, 14, 15, 16, 17, 18, 19, 20],
            [21, 22, 23, 24, 25, 26, 27, 28, 29, 30],
        ]
    )
    b4 = torch.stack([t3, t3, t3, t3])

    _test_eq(b4.shape, torch.Size([4, 3, 10]))
    _test_eq(_shift(t1, 4), torch.tensor([[0, 0, 0, 0, 1, 2, 3, 4, 5, 6]]))
    _test_eq(
        _shift(t3, -2),
        torch.tensor(
            [
                [3, 4, 5, 6, 7, 8, 9, 10, 0, 0],
                [13, 14, 15, 16, 17, 18, 19, 20, 0, 0],
                [23, 24, 25, 26, 27, 28, 29, 30, 0, 0],
            ]
        ),
    )


def test_shift_with_zero():
    _test_eq(_shift(torch.arange(1, 10), 0), torch.arange(1, 10))


def test_shift_invalid_direction(audio):
    with pytest.raises(ValueError):
        SignalShifter(p=1, direction=-2)


def test_shift_max_time(audio):
    shift = SignalShifter(max_time=1)
    inp, out = apply_transform(shift, audio)
    _test_eq(inp.data.shape, out.data.shape)


def test_rolling(audio):
    shift_and_roll = SignalShifter(p=1, max_pct=0.5, roll=True)
    inp, out = apply_transform(shift_and_roll, audio)
    _test_eq(inp.data.shape, out.data.shape)


def test_no_rolling(audio):
    shift_and_roll = SignalShifter(p=1, max_pct=0.5, roll=False)
    inp, out = apply_transform(shift_and_roll, audio)
    _test_eq(inp.data.shape, out.data.shape)


def test_down_mix_mono(audio):
    "Test downmixing 1 channel has no effect"
    downmixer = DownmixMono()
    inp, out = apply_transform(downmixer, audio)
    _test_eq(inp.data, out.data)


def test_noise_fail_bad_color(audio):
    with pytest.raises(ValueError):
        AddNoiseGPU(audio, color=5)

    with pytest.raises(ValueError):
        AddNoiseGPU(audio, color=-3)


def test_noise_white(audio):
    addnoise = AddNoiseGPU(color=NoiseColor.White, p=1.0, min_level=0.1, max_level=0.2)
    inp, out = apply_transform(addnoise, audio)
    _test_ne(inp.data, out.data)


def test_noise_non_white(audio):
    # White noise uses a different method to other noises, so test both.
    addnoise = AddNoiseGPU(color=NoiseColor.Pink, p=1.0, min_level=0.1, max_level=0.2)
    inp, out = apply_transform(addnoise, audio)
    _test_ne(inp.data, out.data)


def test_change_volume(audio):
    changevol = ChangeVolumeGPU(1)
    inp, out = apply_transform(changevol, audio)
    _test_ne(inp.data, out.data)


def test_signal_loss(audio):
    signalloss = SignalLossGPU(1)
    inp, out = apply_transform(signalloss, audio)
    _test_ne(inp.data, out.data)


def test_signal_cutout():
    c, s = 2, 16000
    min_cut_pct, max_cut_pct = 0.10, 0.15
    # Create tensor with no zeros
    audio = AudioTensor(torch.rand([c, s]), sr=16000) * 0.9 + 0.1
    cutout = SignalCutoutGPU(p=1.0, min_cut_pct=min_cut_pct, max_cut_pct=max_cut_pct)
    inp, out = apply_transform(cutout, audio)

    _test_ne(inp.data, out.data)

    num_zeros = (out == 0).sum()
    assert min_cut_pct * s * c <= num_zeros <= max_cut_pct * s * c, num_zeros



import pytest

from fastaudio.augment.spectrogram import (
    CropTime,
    Delta,
    MaskFreq,
    MaskTime,
    SGRoll,
    TfmResize
)
from fastaudio.util import test_audio_tensor


def invoke_class(cls, **args):
    audio = test_audio_tensor()
    with pytest.warns(Warning):
        cls(**args)(audio)


def test_show_warning_with_tfm_on_sig():
    """
    When we invoke a transform intended for a spectrogram on
    something that is most likely a signal, we show a warning.

    Check Issue #17
    """
    transforms = [CropTime, Delta, MaskFreq, MaskTime, SGRoll, TfmResize]

    for t in transforms:
        # Some of the transforms require init arguments
        if t == CropTime:
            invoke_class(t, duration=1)
        elif t == TfmResize:
            invoke_class(t, size=2)
        else:
            invoke_class(t)



import os
from functools import wraps


def _empty():
    return


def skip_if_ci(func):
    @wraps(func)
    def _inner(alt_func=_empty):
        if "CI" in os.environ and os.environ["CI"] == "true":
            return alt_func()
        return func()

    return _inner



import inspect
import math
import torch
from fastai.data.all import test_close as _close
from tempfile import TemporaryFile
from torchaudio.transforms import MelSpectrogram

from fastaudio.all import AudioConfig, AudioTensor, AudioToMFCC, SpectrogramTransformer
from fastaudio.util import test_audio_tensor


# TODO
def test_download_audio():
    pass


def test_load_audio():
    item0 = test_audio_tensor()
    DBMelSpec = SpectrogramTransformer(mel=True, to_db=True)
    a2s = DBMelSpec(f_max=20000, n_mels=137)
    sg = a2s(item0)

    assert type(item0) == AudioTensor
    assert item0.sr == 16000
    assert item0.nchannels == 1
    assert item0.nsamples == 32000
    assert item0.duration == 2

    assert sg.f_max == 20000
    assert sg.hop_length == 512
    assert sg.sr == item0.sr
    assert sg.mel
    assert sg.to_db
    assert sg.nchannels == 1
    assert sg.height == 137
    assert sg.n_mels == sg.height
    assert sg.width == 63

    defaults = {
        k: v.default for k, v in inspect.signature(MelSpectrogram).parameters.items()
    }
    hop_length = 345
    a2s = DBMelSpec(f_max=20000, hop_length=hop_length)
    sg = a2s(item0)
    assert sg.n_mels == defaults["n_mels"]
    assert sg.n_fft == 1024
    assert sg.shape[1] == sg.n_mels
    assert sg.hop_length == hop_length

    # test the spectrogram and audio have same duration, both are computed
    # on the fly as transforms can change their duration
    _close(sg.duration, item0.duration, eps=0.1)


def test_spectrograms_right_side_up():
    DBMelSpec = SpectrogramTransformer(mel=True, to_db=True)
    a2s_5hz = DBMelSpec(
        sample_rate=16000,
        n_fft=1024,
        win_length=1024,
        hop_length=512,
        f_min=0.0,
        f_max=20000,
        pad=0,
        n_mels=137,
    )
    sine_5hz = 0.5 * torch.cos(2 * math.pi * 5 * torch.arange(0, 1.0, 1.0 / 16000))
    at_5hz = AudioTensor(sine_5hz[None], 16000)
    sg_5hz = a2s_5hz(at_5hz)
    max_row = sg_5hz.max(dim=1).indices.mode().values.item()
    assert max_row < 2


def test_add_tensors():
    tst0 = AudioTensor(torch.ones(10), sr=120)
    tst1 = AudioTensor(torch.ones(10), sr=150)
    assert (tst0 + tst1).sr == 120


def test_check_nchannels():
    audio_tensor = AudioTensor(torch.ones(1, 3, 10), sr=120)
    assert audio_tensor.nchannels == 3


def test_saved_audiotensor_keeps_metadata():
    # This test is related to this issue: https://github.com/fastaudio/fastaudio/issues/95
    # What happens is that multiprocessing uses pickling to distribute the data
    # and the way it was done inside fastai breaks when loading the metadata
    audio_tensor = test_audio_tensor()

    with TemporaryFile("wb+") as f:
        torch.save(audio_tensor, f)
        f.seek(0, 0)  # Go back to the begining of the file to read
        new_audio_tensor = torch.load(f)
        assert new_audio_tensor.sr == audio_tensor.sr


def test_saved_spectrogram_keeps_metadata():
    # Same issue as the test above
    item0 = test_audio_tensor()
    DBMelSpec = SpectrogramTransformer(mel=True, to_db=True)
    a2s = DBMelSpec(f_max=20000, n_mels=137)
    sg = a2s(item0)

    with TemporaryFile("wb+") as f:
        torch.save(sg, f)
        f.seek(0, 0)  # Go back to the begining of the file to read
        new_sg = torch.load(f)
        assert new_sg.sr == item0.sr


def test_indexing_audiotensor():
    audio_tensor = test_audio_tensor()
    assert audio_tensor.data[:, :3000].shape[1] == 3000


def test_mfcc_transform():
    audio = test_audio_tensor()
    a2s = AudioToMFCC.from_cfg(AudioConfig.BasicMFCC())
    sg = a2s(audio)
    assert len(sg.shape) == 3


def test_show_spectrogram():
    audio = test_audio_tensor()
    a2s = AudioToMFCC.from_cfg(AudioConfig.BasicMFCC())
    sg = a2s(audio)
    sg.show()



import torchaudio.transforms as transforms
from dataclasses import make_dataclass
from fastai.data.block import TransformBlock
from fastai.data.transforms import IntToFloatTensor
from fastai.imports import Path, partial
from fastcore.meta import delegates
from fastcore.transform import Pipeline
from fastcore.utils import ifnone
from inspect import signature
from torchaudio import save as save_audio

from ..augment.preprocess import Resample
from ..augment.signal import DownmixMono, ResizeSignal
from .signal import AudioTensor, get_audio_files


def audio_item_tfms(sample_rate=16000, force_mono=True, crop_signal_to=None):
    """
    Basic audio preprocessing transforms.
    """
    tfms = []
    if sample_rate is not None:
        tfms.append(Resample(sample_rate))
    if force_mono:
        tfms.append(DownmixMono())
    if crop_signal_to is not None:
        tfms.append(ResizeSignal(duration=crop_signal_to))
    return tfms


class PreprocessAudio:
    """
    Creates an audio tensor and run the basic preprocessing transforms on it.
    Used while preprocessing the audios, this is not a `Transform`.
    """

    @delegates(audio_item_tfms)
    def __init__(self, **kwargs):
        self.tfms = Pipeline(audio_item_tfms(**kwargs))

    def __call__(self, x):
        audio = AudioTensor.create(x)
        return self.tfms(audio)


@delegates(PreprocessAudio, keep=True)
def preprocess_audio_folder(path, folders=None, output_dir=None, **kwargs):
    "Preprocess audio files in `path` in parallel using `n_workers`"
    path = Path(path)
    fnames = get_audio_files(path, recurse=True, folders=folders)
    output_dir = Path(ifnone(output_dir, path.parent / f"{path.name}_cached"))
    output_dir.mkdir(exist_ok=True)

    pp = PreprocessAudio(**kwargs)

    for i, fil in enumerate(fnames):
        out = output_dir / fnames[i].relative_to(path)
        aud = pp(fil)
        save_audio(str(out), aud, aud.sr)
    return output_dir


class AudioBlock(TransformBlock):
    "A `TransformBlock` for audios"

    @delegates(audio_item_tfms)
    def __init__(self, cache_folder=None, **kwargs):
        item_tfms = audio_item_tfms(**kwargs)
        type_tfm = partial(AudioTensor.create, cache_folder=cache_folder)
        return super().__init__(
            type_tfms=type_tfm, item_tfms=item_tfms, batch_tfms=IntToFloatTensor
        )

    @classmethod
    @delegates(audio_item_tfms, keep=True)
    def from_folder(cls, path, **kwargs):
        "Build a `AudioBlock` from a `path` and caches some intermediary results"
        cache_folder = preprocess_audio_folder(path, **kwargs)
        return cls(cache_folder, **kwargs)


def config_from_func(func, name, **kwargs):
    params = signature(func).parameters.items()
    namespace = {k: v.default for k, v in params}
    namespace.update(kwargs)
    func_config = make_dataclass(name, namespace.keys(), namespace=namespace)
    func_config.__doc__ = ""
    return func_config


class AudioConfig:
    """
    Collection of configurations to build `AudioToSpec` transforms.
    """

    # default configurations from the wrapped function
    # make sure to pass in mel=False as kwarg for non-mel spec
    # and to_db=False for non db spec
    BasicSpectrogram = config_from_func(
        transforms.Spectrogram, "BasicSpectrogram", mel=False, to_db=True
    )
    BasicMelSpectrogram = config_from_func(
        transforms.MelSpectrogram, "BasicMelSpectrogram", mel=True, to_db=True
    )
    BasicMFCC = config_from_func(transforms.MFCC, "BasicMFCC ")
    # special configs with domain-specific defaults

    Voice = config_from_func(
        transforms.MelSpectrogram,
        "Voice",
        mel="True",
        to_db="False",
        f_min=50.0,
        f_max=8000.0,
        n_fft=1024,
        n_mels=128,
        hop_length=128,
    )



from fastaudio.util import create_sin_wave, test_audio_tensor


def test_create_sin_wave():
    wave = create_sin_wave()
    assert wave != None
    wave, sr = wave
    assert sr == 16000
    assert wave.shape[0] == 5 * sr


def test_shape_of_sin_wave_tensor():
    sr = 16000
    secs = 2
    ai = test_audio_tensor(secs, sr)
    assert ai.duration == secs
    assert ai.nsamples == secs * sr



import torch
from enum import Enum
from fastai.imports import math, np
from fastcore.transform import Transform
from librosa.effects import split
from scipy.signal import resample_poly

from ..core.signal import AudioTensor


class RemoveType(Enum):
    "All methods of removing silence as attributes to get tab-completion and typo-proofing"
    Trim = "trim"
    All = "all"
    Split = "split"


def _merge_splits(splits, pad):
    clip_end = splits[-1][1]
    merged = []
    i = 0
    while i < len(splits):
        start = splits[i][0]
        while splits[i][1] < clip_end and splits[i][1] + pad >= splits[i + 1][0] - pad:
            i += 1
        end = splits[i][1]
        merged.append(np.array([max(start - pad, 0), min(end + pad, clip_end)]))
        i += 1
    return np.stack(merged)


class RemoveSilence(Transform):
    """Split signal at points of silence greater than 2*pad_ms """

    def __init__(self, remove_type=RemoveType.Trim, threshold=20, pad_ms=20):
        self.remove_type = remove_type
        self.threshold = threshold
        self.pad_ms = pad_ms

    def encodes(self, ai: AudioTensor) -> AudioTensor:
        if self.remove_type is None:
            return ai
        padding = int(self.pad_ms / 1000 * ai.sr)
        if padding > ai.nsamples:
            return ai
        if ai.shape[0] < 2:
            splits = split(ai[0].numpy(), top_db=self.threshold, hop_length=padding)
        else:
            splits = split(ai.numpy(), top_db=self.threshold, hop_length=padding)
        if self.remove_type == RemoveType.Split:
            sig = [
                ai[:, (max(a - padding, 0)) : (min(b + padding, ai.nsamples))]
                for (a, b) in _merge_splits(splits, padding)
            ]
        elif self.remove_type == RemoveType.Trim:
            sig = [ai[:, (max(splits[0, 0] - padding, 0)) : splits[-1, -1] + padding]]
        elif self.remove_type == RemoveType.All:
            sig = [
                torch.cat(
                    [
                        ai[:, (max(a - padding, 0)) : (min(b + padding, ai.nsamples))]
                        for (a, b) in _merge_splits(splits, padding)
                    ],
                    dim=1,
                )
            ]
        else:
            raise ValueError(
                f"""Valid options for silence removal are
                None, RemoveType.Split, RemoveType.Trim, RemoveType.All,
                but not '{self.remove_type}'."""
            )
        ai.data = torch.cat(sig, dim=-1)
        return ai


class Resample(Transform):
    """Resample using faster polyphase technique and avoiding FFT computation"""

    def __init__(self, sr_new):
        self.sr_new = sr_new

    def encodes(self, ai: AudioTensor) -> AudioTensor:
        if ai.sr == self.sr_new:
            return ai
        sig_np = ai.numpy()
        sr_gcd = math.gcd(ai.sr, self.sr_new)
        resampled = resample_poly(
            sig_np, int(self.sr_new / sr_gcd), int(ai.sr / sr_gcd), axis=-1
        )
        ai.data = torch.from_numpy(resampled.astype(np.float32))
        ai.sr = self.sr_new
        return ai



import torch
from fastai.vision.augment import RandTransform
from functools import wraps
from math import pi
from torch import Tensor

from fastaudio.core.signal import AudioTensor

__all__ = ["test_audio_tensor"]


def create_sin_wave(seconds=5, sr=16000, freq=400):
    "Creates a sin wave to be used for tests"
    max_samples = freq * seconds * 2 * pi
    rate = 2 * pi * freq / sr
    samples = torch.arange(0, max_samples, rate, dtype=torch.float)
    sin_wave = torch.sin(samples)
    return sin_wave, sr


def test_audio_tensor(seconds=2, sr=16000, channels=1):
    "Generates an Audio Tensor for testing that is based on a sine wave"
    sin_wave, sr = create_sin_wave(seconds=seconds, sr=sr)
    sin_wave = sin_wave.repeat(channels, 1)
    return AudioTensor(sin_wave, sr)


def apply_transform(transform, inp):
    """Generate a new input, apply transform, and display/return both input and output"""
    inp_orig = inp.clone()
    out = (
        transform(inp_orig, split_idx=0)
        if isinstance(transform, RandTransform)
        else transform(inp_orig)
    )
    return inp.clone(), out


def auto_batch(item_dims):
    """Wrapper that always calls the underlying function with a batch.

    Single items are expanded before calling, then the result is squashed back
    to the original dimensions. Batches are passed without modification. This
    allows the underlying function to be written once (for batches) and
    automatically work on both.

    The overhead of this wrapper is very small and will be dwarfed by most
    operations on reasonably-sized tensors.

    ``item_dims`` specifies the number of dimensions in an item.

    """

    def wrapper(orig_func):
        @wraps(orig_func)
        def expand_and_do(self, x: Tensor):
            nonlocal orig_func, item_dims
            is_item = x.dim() == item_dims
            if is_item:
                # Expand to batch
                x = x.unsqueeze(0)

            x = orig_func(self, x)

            if is_item:
                # Restore original shape
                return x.squeeze(0)
            else:
                return x

        return expand_and_do

    return wrapper
